use deku::ctx::BitSize;
use deku::prelude::*;
use deku::reader::ReaderRet;
use deku::{reader::Reader, DekuError};

pub mod blocks;
pub mod groups;
pub mod prelude;

pub use blocks::Data;

/// # Read the size of a FSV block / field
/// Documentation: SSV Documentation, page 29
fn size_read<R: std::io::Read + std::io::Seek>(reader: &mut Reader<R>) -> Result<u64, DekuError> {
    let size_bytes = u8::from_reader_with_ctx(reader, BitSize(8))?;
    let size: u64 = if size_bytes & 0b1000_0000 == 0 {
        // If the Most Significant Bit is 0, the size is encoded on 7 bits
        size_bytes.into()
    } else {
        // Else, the 7 following bits indicate the number of bytes of the block containing the size
        let size_block_len: usize = (size_bytes & 0b0111_1111).into();
        if size_block_len > 4 {
            return Err(DekuError::Parse(format!("Unexpected size block length: {}", size_block_len).into()));
        };
        // The block containing the size is encoded on 1 to 4 bytes
        let buffer: &mut [u8; 4] = &mut [0; 4];
        let write_offset = 4 - size_block_len;
        match reader.read_bytes(size_block_len, &mut buffer[write_offset..])? {
            ReaderRet::Bits(_bit_vec) => return Err(DekuError::Parse("Unexpected result reading size bytes: got bits".into())),
            ReaderRet::Bytes => u32::from_be_bytes(*buffer).into(),
        }
    };
    Ok(size)
}

/// Deku Reader taking an expected size into account
/// This function limit the reading to the size given in input
fn read_with_size<T, R: std::io::Read + std::io::Seek>(
    reader: &mut Reader<R>,
    size: usize
) -> Result<T, DekuError> 
    where T: for<'a> DekuContainerRead<'a>
{
    let max_size = core::mem::size_of::<T>();
    let mut buf = vec![0; max_size];
    let buf: &mut [u8] = &mut buf;
    let ret = reader.read_bytes(size, buf)?;
    let (_rest, block) = match ret {
        ReaderRet::Bytes => {
            T::from_bytes((buf, 0))?
        },
        _ => return Err(DekuError::Parse("Unexpected result reading size bytes: got bits".into())),
    };
    Ok(block)
}

/// # Map bytes to a lossy string
/// This function is used to map bytes to a string, ignoring invalid UTF-8 characters
/// Example: [0x41, 0x42] -> "AB"
/// Example: [48, 49, 50, 51] -> "0123"
fn map_bytes_to_lossy_string(data: &[u8]) -> Result<String, DekuError> {
    // let data = data.to_vec();
    let version: String = String::from_utf8_lossy(data).to_string();
    Ok(version)
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_read_with_size_reader() {
        #[derive(Debug, PartialEq)]
        #[deku_derive(DekuRead)]
        struct Data {
            #[deku(read_all)]
            pub blocks: Vec<Block>,
        }
        #[derive(Debug, PartialEq)]
        #[deku_derive(DekuRead)]
        struct Block {
            pub size: u8,
            pub id: u8,
            #[deku(ctx = "*id,*size-1")] // size-1 to remove the ID size
            pub data: BlockType,
        }
        #[derive(Debug, PartialEq)]
        #[deku_derive(DekuRead)]
        #[deku(ctx = "id: u8, size: u8", id = "id")]
        enum BlockType {
            #[deku(id = 1)]
            Block1(
                #[deku(reader = "read_with_size(deku::reader, size as usize)")]
                Block1
            ),
            #[deku(id = 2)]
            Block2(
                #[deku(reader = "read_with_size(deku::reader, size as usize)")]
                Block2
            ),
        }
        #[derive(Debug, PartialEq)]
        #[deku_derive(DekuRead)]
        struct Block1 {
            pub field1_size: u8,
            pub field1: u16,
            pub field2_size: u8,
            pub field2: u64,
        }
        #[derive(Debug, PartialEq)]
        #[deku_derive(DekuRead)]
        struct Block2;

        let buffer = &[
            // 1st Block, type 1
            4, // Size: 4
                1, // ID: 1
                    2, 0x12, 0x34, // Field 1, size 2
                    // No Field 2
            // 2nd Block, type 1
            6, // Size: Y
                1, // ID: 2
                    1, 0x56, // Field 1, size 1 (casted into u16)
                    2, 0x78, 0x9A // Field 2, size 2
        ];

        let (_rest, val) = Data::from_bytes((buffer, 0)).unwrap();
        assert_eq!(val.blocks.len(), 2);
        assert_eq!(val.blocks[0].size, 4);
        assert_eq!(val.blocks[1].size, 6);
    }
}